import os
from pathlib import Path
from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
import scvelo as scv
import seaborn as sns
from beartype import beartype

from pyrovelocity.analyze import pareto_frontier_genes
from pyrovelocity.io import CompressedPickle
from pyrovelocity.logging import configure_logging
from pyrovelocity.plots import cluster_violin_plots
from pyrovelocity.plots import extrapolate_prediction_sample_predictive
from pyrovelocity.plots import plot_gene_ranking
from pyrovelocity.plots import plot_parameter_posterior_distributions
from pyrovelocity.plots import posterior_curve
from pyrovelocity.plots import rainbowplot
from pyrovelocity.plots import summarize_fig2_part1
from pyrovelocity.plots import summarize_fig2_part2
from pyrovelocity.utils import save_anndata_counts_to_dataframe


__all__ = ["summarize_dataset"]

logger = configure_logging(__name__)


@beartype
def summarize_dataset(
    data_model: str,
    data_model_path: str | Path,
    model_path: str | Path,
    pyrovelocity_data_path: str | Path,
    postprocessed_data_path: str | Path,
    cell_state: str,
    vector_field_basis: str,
    reports_path: str | Path = "reports",
) -> Tuple[Path, Path]:
    """
    Construct summary plots for each data set and model.

    Args:
        data_model (str): string containing the data set and model identifier,
            e.g. simulated_model1
        model_path (str | Path): path to the model,
            e.g. models/simulated_model1/model
        pyrovelocity_data_path (str | Path): path to the pyrovelocity data,
            e.g. models/simulated_model1/pyrovelocity.pkl.zst
        postprocessed_data_path (str | Path): path to the trained data,
            e.g. models/simulated_model1/postprocessed.h5ad
        cell_state (str): string containing the cell state identifier,
            e.g. cell_type
        vector_field_basis (str): string containing the vector field basis
            identifier, e.g. umap
        reports_path (str | Path): path to the reports, e.g. reports

    Returns:
        Path: Top-level path to reports outputs for the data model combination,
            e.g. reports/simulated_model1

    Examples:
        >>> from pyrovelocity.summarize import summarize_dataset # xdoctest: +SKIP
        >>> tmp = getfixture("tmp_path") # xdoctest: +SKIP
        >>> summarize_dataset(
        ...     "simulated_model1",
        ...     "models/simulated_model1",
        ...     "models/simulated_model1/model",
        ...     "models/simulated_model1/pyrovelocity.pkl.zst",
        ...     "models/simulated_model1/postprocessed.h5ad",
        ...     "leiden",
        ...     "umap",
        ... ) # xdoctest: +SKIP
    """

    logger.info(f"\n\nPlotting summary figure(s) in: {data_model}\n\n")

    data_model_reports_path = Path(reports_path) / Path(data_model)
    logger.info(
        f"\n\nVerifying existence of path for:\n\n"
        f"  reports: {data_model_reports_path}\n"
    )

    Path(data_model_path).mkdir(parents=True, exist_ok=True)
    Path(data_model_reports_path).mkdir(parents=True, exist_ok=True)
    posterior_phase_portraits_path = Path(data_model_reports_path) / Path(
        "posterior_phase_portraits"
    )
    Path(posterior_phase_portraits_path).mkdir(parents=True, exist_ok=True)

    dataframe_path = Path(data_model_path) / f"{data_model}_dataframe.pkl.zst"
    volcano_plot = data_model_reports_path / "volcano.pdf"
    rainbow_plot = data_model_reports_path / "rainbow.pdf"
    vector_field_plot = data_model_reports_path / "vector_field.pdf"
    shared_time_plot = data_model_reports_path / "shared_time.pdf"
    fig2_part1_plot = data_model_reports_path / "fig2_part1_plot.pdf"
    fig2_part2_plot = data_model_reports_path / "fig2_part2_plot.pdf"
    violin_clusters_lin = data_model_reports_path / "violin_clusters_lin.pdf"
    violin_clusters_log = data_model_reports_path / "violin_clusters_log.pdf"
    parameter_uncertainty_plot_path = (
        data_model_reports_path / "parameter_uncertainties.pdf"
    )
    t0_selection_plot = data_model_reports_path / "t0_selection.png"

    output_filenames = [
        dataframe_path,
        volcano_plot,
        rainbow_plot,
        vector_field_plot,
        shared_time_plot,
        parameter_uncertainty_plot_path,
        t0_selection_plot,
    ]
    if all(os.path.isfile(f) for f in output_filenames):
        logger.info(
            "\n\t"
            + "\n\t".join(str(f) for f in output_filenames)
            + "\nAll output files exist"
        )
        logger.warn("Remove output files if you want to regenerate them.")
        return (data_model_reports_path, dataframe_path)

    logger.info(f"Loading trained data: {postprocessed_data_path}")
    adata = scv.read(postprocessed_data_path)
    # gene_mapping = {"1100001G20Rik": "Wfdc21"}
    # adata = rename_anndata_genes(adata, gene_mapping)

    logger.info(f"Loading pyrovelocity data: {pyrovelocity_data_path}")
    posterior_samples = CompressedPickle.load(pyrovelocity_data_path)
    # print(posterior_samples.keys())

    logger.info("Constructing t0 selection plot")
    fig, ax = plt.subplots(5, 6)
    fig.set_size_inches(26, 24)
    ax = ax.flatten()

    posterior_samples_keys_to_check = ["t0", "switching", "cell_time"]

    num_samples_list = [
        posterior_samples[key].shape[0]
        for key in posterior_samples_keys_to_check
    ]
    assert (
        len(set(num_samples_list)) == 1
    ), f"The number of samples is not equal across keys: {posterior_samples_keys_to_check}"

    num_samples = num_samples_list[0]

    for sample in range(num_samples):
        t0_sample = posterior_samples["t0"][sample]
        switching_sample = posterior_samples["switching"][sample]
        cell_time_sample = posterior_samples["cell_time"][sample]
        ax[sample].scatter(
            t0_sample.flatten(),
            2 * np.ones(t0_sample.shape[-1]),
            s=1,
            c="red",
            label="t0",
        )
        ax[sample].scatter(
            switching_sample.flatten(),
            3 * np.ones(t0_sample.shape[-1]),
            s=1,
            c="purple",
            label="switching",
        )
        ax[sample].scatter(
            cell_time_sample.flatten(),
            np.ones(cell_time_sample.shape[0]),
            s=1,
            c="blue",
            label="shared time",
        )
        ax[sample].set_ylim(-0.5, 4)
        if sample == 28:
            ax[sample].legend(loc="best", bbox_to_anchor=(0.5, 0.0, 0.5, 0.5))
        # print((t0_sample.flatten() > cell_time_sample.flatten().max()).sum())
        # print((t0_sample.flatten() < switching_sample.flatten().max()).sum())
        # print((t0_sample.flatten() > switching_sample.flatten().max()).sum())
        # for gene in adata.var_names[
        #     t0_sample.flatten() > cell_time_sample.flatten().max()
        # ]:
        #     print(gene)
    ax[-1].hist(t0_sample.flatten(), bins=200, color="red", alpha=0.3)
    ax[-1].hist(cell_time_sample.flatten(), bins=500, color="blue", alpha=0.3)

    fig.savefig(
        t0_selection_plot,
        facecolor=fig.get_facecolor(),
        bbox_inches="tight",
        edgecolor="none",
        dpi=300,
    )

    logger.info(
        "Extrapolating prediction samples for predictive posterior plots"
    )
    (
        grid_time_samples_ut,
        grid_time_samples_st,
        grid_time_samples_u0,
        grid_time_samples_s0,
        grid_time_samples_uinf,
        grid_time_samples_sinf,
        grid_time_samples_uscale,
        grid_time_samples_state,
        grid_time_samples_t0,
        grid_time_samples_dt_switching,
    ) = extrapolate_prediction_sample_predictive(
        posterior_samples["cell_time"],
        model_path,
        adata,
        grid_time_points=500,
    )
    # extrapolate_prediction_trace(data_model_conf, adata, grid_time_points=5)

    # ##################
    # # save dataframe
    # ##################

    save_anndata_counts_to_dataframe(adata, dataframe_path)

    ##################
    # generate figures
    ##################

    # vector fields
    cell_type = cell_state

    if os.path.isfile(fig2_part1_plot):
        logger.info(f"{fig2_part1_plot} exists")
    else:
        logger.info(f"Generating figure: {fig2_part1_plot}")
        summarize_fig2_part1(
            adata,
            posterior_samples["vector_field_posterior_samples"],
            posterior_samples["cell_time"],
            posterior_samples["original_spaces_embeds_magnitude"],
            posterior_samples["pca_embeds_angle"],
            posterior_samples["embeds_angle"],
            vector_field_basis,
            posterior_samples["vector_field_posterior_mean"],
            cell_type,
            fig2_part1_plot,
        )

    # gene selection
    if os.path.isfile(fig2_part2_plot):
        logger.info(f"{fig2_part2_plot} exists")
    else:
        logger.info(f"Generating figure: {fig2_part2_plot}")
        summarize_fig2_part2(
            adata,
            posterior_samples,
            basis=vector_field_basis,
            cell_state=cell_type,
            plot_name=fig2_part2_plot,
            fig=None,
        )

    # cluster violin plots
    if os.path.isfile(violin_clusters_log):
        logger.info(f"{violin_clusters_log} exists")
    else:
        logger.info(f"Generating figure: {violin_clusters_log}")
        for fig_name in [violin_clusters_lin, violin_clusters_log]:
            cluster_violin_plots(
                data_model,
                adata=adata,
                posterior_samples=posterior_samples,
                cluster_key=cell_state,
                violin_flag=True,
                pairs=None,
                show_outlier=False,
                fig_name=fig_name,
            )

    # shared time plot
    cell_time_mean = posterior_samples["cell_time"].mean(0).flatten()
    cell_time_std = posterior_samples["cell_time"].std(0).flatten()
    adata.obs["shared_time_uncertain"] = cell_time_std
    adata.obs["shared_time_mean"] = cell_time_mean

    if os.path.isfile(shared_time_plot):
        logger.info(f"{shared_time_plot} exists")
    else:
        logger.info(f"Generating figure: {shared_time_plot}")
        fig, ax = plt.subplots(1, 3)
        fig.set_size_inches(9.2, 2.6)
        ax = ax.flatten()
        ax_cb = scv.pl.scatter(
            adata,
            c="shared_time_mean",
            ax=ax[0],
            show=False,
            cmap="inferno",
            fontsize=7,
            colorbar=True,
        )
        ax_cb = scv.pl.scatter(
            adata,
            c="shared_time_uncertain",
            ax=ax[1],
            show=False,
            cmap="inferno",
            fontsize=7,
            colorbar=True,
        )
        select = adata.obs["shared_time_uncertain"] > np.quantile(
            adata.obs["shared_time_uncertain"], 0.9
        )
        sns.kdeplot(
            x=adata.obsm[f"X_{vector_field_basis}"][:, 0][select],
            y=adata.obsm[f"X_{vector_field_basis}"][:, 1][select],
            ax=ax[1],
            levels=3,
            fill=False,
        )
        ax[2].hist(cell_time_std / cell_time_mean, bins=100)
        fig.savefig(
            shared_time_plot,
            facecolor=fig.get_facecolor(),
            bbox_inches="tight",
            edgecolor="none",
            dpi=300,
        )

    volcano_data = posterior_samples["gene_ranking"]
    number_of_marker_genes = min(
        max(int(len(volcano_data) * 0.1), 4), 20, len(volcano_data)
    )
    logger.info(f"Searching for {number_of_marker_genes} marker genes")
    geneset = pareto_frontier_genes(volcano_data, number_of_marker_genes)

    logger.info("Generating posterior phase portraits")
    posterior_curve(
        adata,
        posterior_samples,
        grid_time_samples_ut,
        grid_time_samples_st,
        grid_time_samples_u0,
        grid_time_samples_s0,
        grid_time_samples_uinf,
        grid_time_samples_sinf,
        grid_time_samples_uscale,
        grid_time_samples_state,
        grid_time_samples_t0,
        grid_time_samples_dt_switching,
        geneset,
        data_model,
        posterior_phase_portraits_path,
    )

    # volcano plot
    if os.path.isfile(volcano_plot):
        logger.info(f"{volcano_plot} exists")
    else:
        logger.info(f"Generating figure: {volcano_plot}")

        volcano_data, fig = plot_gene_ranking(
            [posterior_samples],
            [adata],
            selected_genes=geneset,
            time_correlation_with="st",
            show_marginal_histograms=True,
        )

        fig.subplots_adjust(wspace=0.1, hspace=0.1)
        for ext in ["", ".png"]:
            fig.savefig(
                f"{volcano_plot}{ext}",
                facecolor=fig.get_facecolor(),
                bbox_inches="tight",
                edgecolor="none",
                dpi=300,
            )

    # parameter uncertainty
    if os.path.isfile(parameter_uncertainty_plot_path):
        logger.info(f"{parameter_uncertainty_plot_path} exists")
    else:
        logger.info(f"Generating figure: {parameter_uncertainty_plot_path}")
        plot_parameter_posterior_distributions(
            posterior_samples=posterior_samples,
            adata=adata,
            geneset=geneset,
            parameter_uncertainty_plot_path=parameter_uncertainty_plot_path,
        )

    # rainbow plot

    if os.path.isfile(rainbow_plot):
        logger.info(f"{rainbow_plot} exists")
    else:
        logger.info(f"Generating figure: {rainbow_plot}")
        fig = rainbowplot(
            volcano_data=volcano_data,
            adata=adata,
            posterior_samples=posterior_samples,
            genes=geneset,
            data=["st", "ut"],
            basis=vector_field_basis,
            cell_state=cell_state,
        )

        for ext in ["", ".png"]:
            fig.savefig(
                f"{rainbow_plot}{ext}",
                facecolor=fig.get_facecolor(),
                bbox_inches="tight",
                edgecolor="none",
                dpi=300,
            )

    # mean vector field plot
    if os.path.isfile(vector_field_plot):
        logger.info(f"{vector_field_plot} exists")
    else:
        logger.info(f"Generating figure: {vector_field_plot}")
        fig, ax = plt.subplots()

        scv.pl.velocity_embedding_grid(
            adata,
            basis=vector_field_basis,
            color=cell_state,
            title="",
            vkey="velocity_pyro",
            linewidth=1,
            ax=ax,
            show=False,
            legend_loc="right margin",
            density=0.4,
            scale=0.2,
            arrow_size=2,
            arrow_length=2,
            arrow_color="black",
        )
        for ext in ["", ".png"]:
            fig.savefig(
                f"{vector_field_plot}{ext}",
                facecolor=fig.get_facecolor(),
                bbox_inches="tight",
                edgecolor="none",
                dpi=300,
            )

    return (data_model_reports_path, dataframe_path)



import matplotlib.pyplot as plt

from pyrovelocity.logging import configure_logging
from pyrovelocity.plots._genes import plot_gene_ranking
from pyrovelocity.plots._rainbow import rainbowplot
from pyrovelocity.plots._time import plot_posterior_time


logger = configure_logging(__name__)


__all__ = ["summarize_fig2_part2"]


def summarize_fig2_part2(
    adata, posterior_samples, plot_name="", basis="", cell_state="", fig=None
):
    if fig is None:
        fig = plt.figure(figsize=(9.5, 5))
        subfigs = fig.subfigures(
            1, 2, wspace=0.0, hspace=0, width_ratios=[1.8, 4]
        )
        ax = subfigs[0].subplots(2, 1)
        plot_posterior_time(
            posterior_samples,
            adata,
            ax=ax[0],
            fig=subfigs[0],
            addition=False,
            basis=basis,
        )
        volcano_data, _ = plot_gene_ranking(
            [posterior_samples], [adata], ax=ax[1], time_correlation_with="st"
        )
        _ = rainbowplot(
            volcano_data,
            adata,
            posterior_samples,
            subfigs[1],
            data=["st", "ut"],
            basis=basis,
            cell_state=cell_state,
            num_genes=4,
        )
        for ext in ["", ".png"]:
            fig.savefig(
                f"{plot_name}{ext}",
                facecolor=fig.get_facecolor(),
                bbox_inches="tight",
                edgecolor="none",
                dpi=300,
            )



import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from adjustText import adjust_text
from matplotlib import gridspec

from pyrovelocity.analyze import compute_volcano_data
from pyrovelocity.logging import configure_logging


__all__ = ["plot_gene_ranking"]

logger = configure_logging(__name__)


def plot_gene_ranking(
    posterior_samples,
    adata,
    ax=None,
    time_correlation_with="s",
    selected_genes=None,
    assemble=False,
    data="correlation",
    negative=False,
    adjust_text_bool=False,
    show_marginal_histograms=False,
) -> None:
    if selected_genes is not None:
        assert isinstance(selected_genes, (tuple, list))
        assert isinstance(selected_genes[0], str)
        volcano_data = posterior_samples[0]["gene_ranking"]
        genes = selected_genes
    elif "u" in posterior_samples[0]:
        volcano_data, genes = compute_volcano_data(
            posterior_samples,
            adata,
            time_correlation_with,
            selected_genes,
            negative,
        )
    else:
        volcano_data = posterior_samples[0]["gene_ranking"]
        genes = posterior_samples[0]["genes"]

    adjust_text_compatible = is_adjust_text_compatible()
    fig = None

    if data == "correlation":
        defaultfontsize = 7
        defaultdotsize = 3
        plot_title = "Pyro-Velocity genes"

        if show_marginal_histograms:
            time_corr_hist, time_corr_bins = np.histogram(
                volcano_data["time_correlation"], bins="auto", density=False
            )
            mean_mae_hist, mean_mae_bins = np.histogram(
                volcano_data["mean_mae"], bins="auto", density=False
            )

            fig = plt.figure(figsize=(10, 10))
            # ax_scatter = plt.subplot2grid((3, 3), (1, 0), colspan=2, rowspan=2)
            # ax_hist_x = plt.subplot2grid((3, 3), (0, 0), colspan=2)
            # ax_hist_y = plt.subplot2grid((3, 3), (1, 2), rowspan=2)
            gs = gridspec.GridSpec(
                3, 3, width_ratios=[2, 2, 1], height_ratios=[1, 2, 2]
            )
            ax_scatter = plt.subplot(gs[1:, :2])
            ax_hist_x = plt.subplot(gs[0, :2])
            ax_hist_y = plt.subplot(gs[1:, 2])

            # time histogram
            ax_hist_x.bar(
                time_corr_bins[:-1],
                time_corr_hist,
                width=np.diff(time_corr_bins),
                align="edge",
            )

            # MAE histogram
            ax_hist_y.barh(
                mean_mae_bins[:-1],
                mean_mae_hist,
                height=np.diff(mean_mae_bins),
                align="edge",
            )
            ax_hist_x.tick_params(axis="x", labelbottom=False)
            ax_hist_y.tick_params(axis="y", labelleft=False)

            defaultfontsize = 14
            defaultdotsize = 12
            plot_title = ""
            ax = ax_scatter

        sns.scatterplot(
            x="time_correlation",
            y="mean_mae",
            hue="selected genes",
            data=volcano_data,
            s=defaultdotsize,
            linewidth=0,
            ax=ax,
            legend=False,
            alpha=0.3,
        )
        ax.set_title(plot_title, fontsize=defaultfontsize)
        ax.set_xlabel(
            "shared time correlation\nwith spliced expression",
            fontsize=defaultfontsize,
        )
        ax.set_ylabel("negative mean\nabsolute error", fontsize=defaultfontsize)
        sns.despine()
        ax.tick_params(labelsize=defaultfontsize - 1)

        texts = []
        for i, g in enumerate(genes):
            ax.scatter(
                volcano_data.loc[g, :].time_correlation,
                volcano_data.loc[g, :].mean_mae,
                s=15,
                color="red",
                marker="*",
            )
            texts.append(
                ax.text(
                    volcano_data.loc[g, :].time_correlation,
                    volcano_data.loc[g, :].mean_mae,
                    g,
                    fontsize=defaultfontsize - 2,
                    color="black",
                    ha="center",
                    va="center",
                )
            )
            if not assemble and adjust_text_compatible:
                # TODO: remove unused code in plot_gene_ranking
                # if i % 2 == 0:
                #     offset = 10 + i * 5
                # else:
                #     offset = -10 - i * 5
                # if i % 2 == 0:
                #     offset_y = -10 + i * 5
                # else:
                #     offset_y = -10 + i * 5
                if not adjust_text_bool:
                    adjust_text(
                        texts,
                        arrowprops=dict(arrowstyle="-", color="red", alpha=0.5),
                        ha="center",
                        va="bottom",
                        ax=ax,
                    )
                else:
                    adjust_text(
                        texts,
                        precision=0.001,
                        expand_text=(1.01, 1.05),
                        expand_points=(1.01, 1.05),
                        force_text=(0.01, 0.25),
                        force_points=(0.01, 0.25),
                        arrowprops=dict(
                            arrowstyle="-", color="blue", alpha=0.6
                        ),
                        ax=ax,
                    )
    else:
        sns.scatterplot(
            x="switching",
            y="mean_mae",
            hue="selected genes",
            data=volcano_data,
            s=3,
            linewidth=0,
            ax=ax,
            legend=False,
            alpha=0.3,
        )
        ax.set_title("Pyro-Velocity genes", fontsize=7)
        ax.set_xlabel("gene switching time", fontsize=7)
        ax.set_ylabel("negative mean\nabsolute error", fontsize=7)
        sns.despine()

    return volcano_data, fig


def is_adjust_text_compatible():
    """Check if the current backend supports adjust_text."""
    try:
        plt.figure()
        test_text = plt.text(0.5, 0.5, "test")
        plt.close()
        adjust_text([test_text], autoalign="y")
        return True
    except Exception as e:
        logger.warning(
            f"adjust_text may not be compatible with the current backend: {e}"
        )
        return False



import anndata
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from pyrovelocity.logging import configure_logging


__all__ = ["plot_parameter_posterior_distributions"]

logger = configure_logging(__name__)


def plot_parameter_posterior_distributions(
    posterior_samples,
    adata: anndata.AnnData,
    geneset,
    parameter_uncertainty_plot_path: str,
):
    fig, ax = plt.subplots(3, 1)
    fig.set_size_inches(18, 12)
    for index, kinetics in enumerate(["alpha", "beta", "gamma"]):
        df = pd.DataFrame(
            np.log(
                posterior_samples[kinetics].squeeze()[
                    :, np.isin(adata.var_names, list(geneset))
                ],
            ),
            columns=adata.var_names[np.isin(adata.var_names, list(geneset))],
        )
        df = df.apply(lambda x: x - x.mean())
        df_long = df.melt(var_name="index", value_name="value")
        logger.debug(df_long.head())
        df_long["index"] = pd.Categorical(
            df_long["index"], categories=geneset, ordered=True
        )

        can_cast_all = (
            df_long["index"].astype(str).apply(lambda x: x.isdigit()).all()
        )
        if can_cast_all:
            df_long["index"] = (
                df_long["index"].astype(str).apply(lambda x: f"g_{int(x)}")
            )
            logger.warning(
                f"Converted integer index to string 'g_int' for {kinetics}"
            )
        else:
            pass

        ax1 = sns.violinplot(
            x="index",
            y="value",
            data=df_long,
            ax=ax[index],
        )
        ax1.set_xticklabels(ax1.get_xticklabels(), rotation=30, ha="right")
        ax1.set_ylabel(kinetics)
        ax1.set_xlabel("")
    fig.subplots_adjust(
        hspace=0.4, wspace=0.45, left=0.08, right=0.95, top=0.9, bottom=0.15
    )
    for ext in ["", ".png"]:
        fig.savefig(
            f"{parameter_uncertainty_plot_path}{ext}",
            facecolor=fig.get_facecolor(),
            bbox_inches="tight",
            edgecolor="none",
            dpi=300,
        )



import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
from scipy.stats import spearmanr

from pyrovelocity.plots._common import set_colorbar


__all__ = ["plot_posterior_time"]


def plot_posterior_time(
    posterior_samples,
    adata,
    ax=None,
    fig=None,
    basis="umap",
    addition=True,
    position="left",
    s=3,
):
    if addition:
        sns.set_style("white")
        sns.set_context("paper", font_scale=1)
        matplotlib.rcParams.update({"font.size": 7})
        plt.figure()
        plt.hist(posterior_samples["cell_time"].mean(0), bins=100, label="test")
        plt.xlabel("mean of cell time")
        plt.ylabel("frequency")
        plt.title("Histogram of cell time posterior samples")
        plt.legend()
    pos_mean_time = posterior_samples["cell_time"].mean(0)
    adata.obs["cell_time"] = pos_mean_time / pos_mean_time.max()

    if ax is None:
        fig, ax = plt.subplots(1, 1)
        fig.set_size_inches(2.36, 2)
    im = ax.scatter(
        adata.obsm[f"X_{basis}"][:, 0],
        adata.obsm[f"X_{basis}"][:, 1],
        s=s,
        alpha=0.4,
        c=adata.obs["cell_time"],
        cmap="inferno",
        linewidth=0,
    )
    set_colorbar(im, ax, labelsize=5, fig=fig, position=position)
    ax.axis("off")
    if "cytotrace" in adata.obs.columns:
        ax.set_title(
            "Pyro-Velocity shared time\ncorrelation with Cytotrace: %.2f"
            % (
                spearmanr(
                    adata.obs["cell_time"].values,
                    1 - adata.obs.cytotrace.values,
                )[0]
            ),
            fontsize=7,
        )
    else:
        ax.set_title("Pyro-Velocity shared time\n", fontsize=7)



from typing import Dict
from typing import List

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scvelo as scv
import seaborn as sns
from anndata import AnnData
from matplotlib import ticker
from matplotlib.figure import Figure
from numpy import ndarray

from pyrovelocity.plots._common import set_colorbar


__all__ = ["rainbowplot", "us_rainbowplot"]


# def rainbowplot(
#     volcano_data,
#     adata,
#     posterior_samples,
#     fig=None,
#     genes=None,
#     data=["st", "ut"],
#     cell_state="clusters",
#     basis="umap",
#     num_genes=5,
#     add_line=True,
#     negative=False,
#     scvelo_colors=False,
# ) -> None:
#     set_font_size(7)

#     if genes is None:
#         genes = get_genes(volcano_data, num_genes, negative)
#     number_of_genes = len(genes)

#     subplot_height = 1  # Set the height of each subplot
#     subplot_width = (
#         subplot_height * 2.0 * 3
#     )  # Width for each row of subplots to keep 1:1 aspect ratio

#     if fig is None:
#         fig, ax = plt.subplots(
#             number_of_genes,
#             3,
#             figsize=(subplot_width, subplot_height * number_of_genes),
#         )

#     if scvelo_colors:
#         colors = setup_scvelo_colors(adata, cell_state)
#     else:
#         colors = setup_colors(adata, cell_state)

#     st, ut = get_posterior_samples(data, posterior_samples)

#     for n, gene in enumerate(genes):
#         ress = get_data(gene, st, ut, adata, cell_state, posterior_samples)
#         ax1 = ax[n, 1]
#         ax2 = ax[n, 0]
#         ax3 = ax[n, 2]
#         if n == 0:
#             ax1.set_title("Rainbow plot", fontsize=7)
#             ax2.set_title("Phase portrait", fontsize=7)
#             ax3.set_title("Denoised spliced", fontsize=7)
#         plot_gene(ax1, ress, colors, add_line)
#         scatterplot(ax2, ress, colors)
#         (index,) = np.where(adata.var_names == gene)
#         im = ax3.scatter(
#             adata.obsm[f"X_{basis}"][:, 0],
#             adata.obsm[f"X_{basis}"][:, 1],
#             s=3,
#             c=st[:, index].flatten(),
#             cmap="RdBu_r",
#         )
#         set_colorbar(im, ax3, labelsize=5, fig=fig, rainbow=True)
#         ax3.axis("off")
#         set_labels(ax1, ax2, ax3, gene, number_of_genes, ress, n)

#     sns.despine()
#     fig.subplots_adjust(
#         hspace=0.5,
#         wspace=0.5,  # left=0.15, right=0.4, top=0.92, bottom=0.12
#     )
#     return fig


# def set_subfigure_titles(ax, n):
#     if n == 0:
#         ax[0].set_title("Rainbow plot", fontsize=7)
#         ax[1].set_title("Phase portrait", fontsize=7)


# def plot_gene_data(ax, ress, colors, add_line):
#     plot_gene(ax[1], ress, colors, add_line)
#     scatterplot(ax[0], ress, colors)


# def scatter_gene(ax, adata, st, gene, basis):
#     (index,) = np.where(adata.var_names == gene)
#     return ax.scatter(
#         adata.obsm[f"X_{basis}"][:, 0],
#         adata.obsm[f"X_{basis}"][:, 1],
#         s=3,
#         c=st[:, index].flatten(),
#         cmap="RdBu_r",
#     )


# def adjust_subfigure(subfig):
#     subfig[0].subplots_adjust(
#         hspace=0.8, wspace=1.4, left=0.32, right=0.94, top=0.92, bottom=0.12
#     )
#     subfig[1].subplots_adjust(
#         hspace=0.8, wspace=0.4, left=0.2, right=0.7, top=0.92, bottom=0.08
#     )

#     subfig[0].text(
#         -0.025,
#         0.58,
#         "unspliced expression",
#         size=7,
#         rotation="vertical",
#         va="center",
#     )
#     subfig[0].text(
#         0.552,
#         0.58,
#         "spliced expression",
#         size=7,
#         rotation="vertical",
#         va="center",
#     )

#     sns.despine()


# def set_font_size(size: int):
#     matplotlib.rcParams.update({"font.size": size})


# def get_genes(volcano_data, num_genes, negative):
#     return (
#         volcano_data.sort_values("mean_mae", ascending=False)
#         .head(300)
#         .sort_values("time_correlation", ascending=negative)
#         .head(num_genes)
#         .index
#     )


# def setup_scvelo_colors(adata, cell_state, basis):
#     scv.pl.scatter(
#         adata,
#         basis=basis,
#         fontsize=7,
#         legend_loc="on data",
#         legend_fontsize=7,
#         color=cell_state,
#         show=False,
#     )
#     return dict(
#         zip(adata.obs.state_info.cat.categories, adata.uns["state_info_colors"])
#     )


# def setup_colors(adata, cell_state):
#     clusters = adata.obs.loc[:, cell_state]
#     return dict(
#         zip(
#             clusters.cat.categories,
#             sns.color_palette("deep", clusters.cat.categories.shape[0]),
#         )
#     )


# def get_posterior_samples(data, posterior_samples):
#     if (data[0] in posterior_samples) and (data[1] in posterior_samples):
#         st = posterior_samples[data[0]].mean(0).squeeze()
#         ut = posterior_samples[data[1]].mean(0).squeeze()
#     else:
#         st = posterior_samples["st_mean"]
#         ut = posterior_samples["ut_mean"]
#     return st, ut


# def get_data(gene, st, ut, adata, cell_state, posterior_samples):
#     (index,) = np.where(adata.var_names == gene)
#     pos_mean_time = posterior_samples["cell_time"].mean(0).flatten()
#     ress = pd.DataFrame(
#         {
#             "cell_time": pos_mean_time / pos_mean_time.max(),
#             "cell_type": adata.obs[cell_state].values,
#             "spliced": st[:, index].flatten(),
#             "unspliced": ut[:, index].flatten(),
#         }
#     )
#     return ress.sort_values("cell_time")


# def plot_gene(ax1, ress, colors, add_line):
#     sns.scatterplot(
#         x="cell_time",
#         y="spliced",
#         data=ress,
#         alpha=0.1,
#         linewidth=0,
#         edgecolor="none",
#         hue="cell_type",
#         palette=colors,
#         ax=ax1,
#         marker="o",
#         legend=False,
#         s=5,
#     )
#     if add_line:
#         for row in range(ress.shape[0]):
#             ax1.vlines(
#                 x=ress.cell_time[row],
#                 ymin=0,
#                 ymax=ress.spliced[row],
#                 colors=colors[ress.cell_type[row]],
#                 alpha=0.1,
#             )


# def set_labels(ax1, ax2, ax3, gene, ngenes, ress, n):
#     if n == 0:
#         ax3.set_title("Denoised spliced", fontsize=7)
#     if n == ngenes - 1:
#         ax1.set_xlabel("shared time", fontsize=7)
#         ax1.set_ylabel("expression", fontsize=7)
#         ax2.set_xlabel("spliced", fontsize=7)
#         ax2.set_ylabel("uspliced", fontsize=7)
#     else:
#         ax1.set_xlabel("")
#         ax2.set_xlabel("")
#         ax1.set_ylabel("")
#         ax2.set_ylabel("")
#     ax2.text(
#         -0.5,
#         0.5,
#         gene,
#         fontsize=8,
#         weight="normal",
#         rotation=0,
#         va="center",
#         ha="right",
#         transform=ax2.transAxes,
#     )

#     x_ticks = [0, np.power(10, get_closest_pow_of_10(ress["cell_time"].max()))]
#     y_ticks = [0, np.power(10, get_closest_pow_of_10(ress["spliced"].max()))]
#     ax1.set_xticks(x_ticks)
#     ax1.set_yticks(y_ticks)
#     ax1.get_xaxis().set_major_formatter(
#         ticker.FuncFormatter(lambda x, _: construct_log_string(x))
#     )
#     ax1.get_yaxis().set_major_formatter(
#         ticker.FuncFormatter(lambda x, _: construct_log_string(x))
#     )
#     ax1.tick_params(axis="both", which="major", labelsize=7)

#     x_ticks = [0, np.power(10, get_closest_pow_of_10(ress["spliced"].max()))]
#     y_ticks = [0, np.power(10, get_closest_pow_of_10(ress["unspliced"].max()))]
#     ax2.set_xticks(x_ticks)
#     ax2.set_yticks(y_ticks)
#     ax2.get_xaxis().set_major_formatter(
#         ticker.FuncFormatter(lambda x, _: construct_log_string(x))
#     )
#     ax2.get_yaxis().set_major_formatter(
#         ticker.FuncFormatter(lambda x, _: construct_log_string(x))
#     )
#     ax2.tick_params(axis="both", which="major", labelsize=7)


# def get_closest_pow_of_10(value):
#     return np.floor(np.log10(value))


# def construct_log_string(x):
#     str_val = f"$10^{{{int(np.log10(x))}}}$" if x > 0 else "0"
#     return rf"{str_val}"


# def scatterplot(ax2, ress, colors):
#     sns.scatterplot(
#         x="spliced",
#         y="unspliced",
#         data=ress,
#         alpha=0.4,
#         linewidth=0,
#         edgecolor="none",
#         hue="cell_type",
#         palette=colors,
#         ax=ax2,
#         marker="o",
#         legend=False,
#         s=3,
#     )


def us_rainbowplot(
    genes: pd.Index,
    adata: AnnData,
    posterior_samples: Dict[str, ndarray],
    data: List[str] = ["st", "ut"],
    cell_state: str = "clusters",
) -> Figure:
    import matplotlib.lines as mlines

    fig, ax = plt.subplots(len(genes), 2)
    fig.set_size_inches(7, 14)
    n = 0
    if data[0] in posterior_samples:
        pos_s = posterior_samples[data[0]].mean(0).squeeze()
        pos_u = posterior_samples[data[1]].mean(0).squeeze()
    else:
        pos_u = posterior_samples["ut_mean"]
        pos_s = posterior_samples["st_mean"]

    for gene in genes:
        (index,) = np.where(adata.var_names == gene)
        ax1 = ax[n, 1]
        if n == 0:
            ax1.set_title("Rainbow plot")
        ress = pd.DataFrame(
            {
                "cell_time": posterior_samples["cell_time"].mean(0).squeeze(),
                "cell_type": adata.obs[cell_state].values,
                "spliced": pos_s[:, index].squeeze(),
                "unspliced": pos_u[:, index].squeeze(),
            }
        )
        if n == 2:
            sns.scatterplot(
                x="cell_time",
                y="spliced",
                data=ress,
                alpha=0.4,
                linewidth=0,
                edgecolor="none",
                hue="cell_type",
                ax=ax1,
                marker="o",
                legend="brief",
                palette="bright",
                s=10,
            )
        else:
            sns.scatterplot(
                x="cell_time",
                y="spliced",
                data=ress,
                alpha=0.4,
                linewidth=0,
                edgecolor="none",
                palette="bright",
                hue="cell_type",
                ax=ax1,
                marker="o",
                legend="brief",
                s=10,
            )

        ax2 = ax[n, 0]
        ax2.set_title(gene)
        ax2.set_ylabel("")
        ax2.set_xlabel("")
        sns.scatterplot(
            x="spliced",
            y="unspliced",
            data=ress,
            alpha=0.4,
            s=25,
            edgecolor="none",
            hue="cell_type",
            ax=ax2,
            legend=False,
            marker="*",
            palette="bright",
        )
        if n == 3:
            blue_star = mlines.Line2D(
                [],
                [],
                color="black",
                marker="o",
                linestyle="None",
                markersize=5,
                label="Spliced",
            )
            red_square = mlines.Line2D(
                [],
                [],
                color="black",
                marker="+",
                linestyle="None",
                markersize=5,
                label="Unspliced",
            )
            ax1.legend(
                handles=[blue_star, red_square], bbox_to_anchor=[2, -0.03]
            )
            ax1.set_xlabel("")
        n += 1
        ax1.legend(bbox_to_anchor=[2, 0.1])
        ax1.tick_params(labelbottom=True)
        ax2.set_xlabel("spliced")
        ax2.set_title(gene)

    plt.subplots_adjust(hspace=0.8, wspace=0.6, left=0.1, right=0.91)
    return fig


def rainbowplot(
    volcano_data,
    adata,
    posterior_samples,
    fig=None,
    genes=None,
    data=["st", "ut"],
    cell_state="clusters",
    basis="umap",
    num_genes=5,
    add_line=True,
    negative=False,
    scvelo_colors=False,
) -> None:
    matplotlib.rcParams.update({"font.size": 7})
    if genes is None:
        genes = (
            volcano_data.sort_values("mean_mae", ascending=False)
            .head(300)
            .sort_values("time_correlation", ascending=negative)
            .head(num_genes)
            .index
        )
    if fig is None:
        fig = plt.figure(figsize=(5.5, 4.5))

    if scvelo_colors:
        scv.pl.scatter(
            adata,
            basis=basis,
            fontsize=7,
            legend_loc="on data",
            legend_fontsize=7,
            color=cell_state,
            show=False,
        )
        colors = dict(
            zip(
                adata.obs.state_info.cat.categories,
                adata.uns["state_info_colors"],
            )
        )
    else:
        clusters = adata.obs.loc[:, cell_state]
        colors = dict(
            zip(
                clusters.cat.categories,
                sns.color_palette("deep", clusters.cat.categories.shape[0]),
            )
        )

    subfigs = fig.subfigures(1, 2, wspace=0.0, width_ratios=[3, 1.5])

    ax = subfigs[0].subplots(len(genes), 2)
    ax_fig2 = subfigs[1].subplots(len(genes), 1)

    n = 0

    if (data[0] in posterior_samples) and (data[1] in posterior_samples):
        st = posterior_samples[data[0]].mean(0).squeeze()
        ut = posterior_samples[data[1]].mean(0).squeeze()
    else:
        st = posterior_samples["st_mean"]
        ut = posterior_samples["ut_mean"]

    for gene in genes:
        (index,) = np.where(adata.var_names == gene)
        ax1 = ax[n, 1]
        ax2 = ax[n, 0]
        ax3 = ax_fig2[n]
        if n == 0:
            ax1.set_title("Rainbow plot", fontsize=7)
            ax2.set_title("Phase portrait", fontsize=7)
        pos_mean_time = posterior_samples["cell_time"].mean(0).flatten()
        ress = pd.DataFrame(
            {
                "cell_time": pos_mean_time / pos_mean_time.max(),
                "cell_type": adata.obs[cell_state].values,
                "spliced": st[:, index].flatten(),
                "unspliced": ut[:, index].flatten(),
            }
        )
        sns.scatterplot(
            x="cell_time",
            y="spliced",
            data=ress,
            alpha=0.1,
            linewidth=0,
            edgecolor="none",
            hue="cell_type",
            palette=colors,
            ax=ax1,
            marker="o",
            legend=False,
            s=5,
        )
        if add_line:
            ress = ress.sort_values("cell_time")
            for row in range(ress.shape[0]):
                ax1.vlines(
                    x=ress.cell_time[row],
                    ymin=0,
                    ymax=ress.spliced[row],
                    colors=colors[ress.cell_type[row]],
                    alpha=0.1,
                )

        if n == len(genes) - 1:
            ax1.set_xlabel("shared time", fontsize=7)
        else:
            ax1.set_xlabel("")
        ax1.set_ylabel("")
        t = [0, round(ress["cell_time"].max(), 5)]
        t_label = ["0", "%.1E" % ress["cell_time"].max()]
        ax1.set_xticks(t, t_label, fontsize=7)
        t = [0, round(ress["spliced"].max(), 5)]
        t_label = ["0", "%.1E" % ress["spliced"].max()]
        ax1.set_yticks(t, t_label, fontsize=7)
        ress = pd.DataFrame(
            {
                "cell_type": adata.obs[cell_state].values,
                "unspliced": ut[:, index].flatten(),
                "spliced": st[:, index].flatten(),
            }
        )
        sns.scatterplot(
            x="spliced",
            y="unspliced",
            data=ress,
            alpha=0.4,
            linewidth=0,
            edgecolor="none",
            hue="cell_type",
            palette=colors,
            ax=ax2,
            marker="o",
            legend=False,
            s=3,
        )
        ax2.set_xlabel("")
        ax2.set_ylabel(gene, fontsize=7, rotation=0, labelpad=23)
        if n == len(genes) - 1:
            ax2.set_xlabel("spliced", fontsize=7)
        t = [0, round(ress["unspliced"].max(), 5)]
        t_label = ["0", "%.1E" % ress["unspliced"].max()]
        ax2.set_yticks(t, t_label, fontsize=7)
        t = [0, round(ress["spliced"].max(), 5)]
        t_label = ["0", "%.1E" % ress["spliced"].max()]
        ax2.set_xticks(t, t_label, fontsize=7)
        im = ax3.scatter(
            adata.obsm[f"X_{basis}"][:, 0],
            adata.obsm[f"X_{basis}"][:, 1],
            s=3,
            c=st[:, index].flatten(),
            cmap="RdBu_r",
        )
        set_colorbar(im, ax3, labelsize=5, fig=subfigs[1], rainbow=True)
        ax3.axis("off")
        if n == 0:
            ax3.set_title("Denoised spliced", fontsize=7)
        n += 1
    sns.despine()
    subfigs[0].subplots_adjust(
        hspace=0.8, wspace=1.4, left=0.32, right=0.94, top=0.92, bottom=0.12
    )
    subfigs[1].subplots_adjust(
        hspace=0.8, wspace=0.4, left=0.2, right=0.7, top=0.92, bottom=0.08
    )
    subfigs[0].text(
        -0.025,
        0.58,
        "unspliced expression",
        size=7,
        rotation="vertical",
        va="center",
    )
    subfigs[0].text(
        0.552,
        0.58,
        "spliced expression",
        size=7,
        rotation="vertical",
        va="center",
    )
    return fig
